import abc
from abc import ABC
from typing import Any, Type, Union

import pyarrow  # type: ignore
from pydantic import BaseModel


class ObjectSerializer(abc.ABC):
    @classmethod
    @abc.abstractmethod
    def serialize(
        cls, value: Any, output_model: Type[BaseModel] = None
    ) -> Union[pyarrow.Table, BaseModel]:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, value: Any) -> Any:
        raise NotImplementedError


class CustomDeserializingModelT(ABC, BaseModel):
    def deserialize(self):
        ...



import abc
import typing as t

import numpy as np
import pandas as pd  # type: ignore
from fastapi import Response
from pydantic import BaseModel

from tktl.core.future.t import EndpointKinds, ProfileKinds


class Endpoint(abc.ABC):
    kind: EndpointKinds

    def __init__(
        self,
        name: str,
        func: t.Callable[..., t.Coroutine[t.Any, t.Any, Response]],
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None] = None,
        profile_columns: t.Optional[t.List[str]] = None,
        profile: t.Optional[ProfileKinds] = None,
        **kwargs,
    ):
        self._kwargs = kwargs
        self._name = name
        self._func = func
        self._X = X
        self._y = y
        self._profile_columns = profile_columns
        self._profile_kind = profile

    @property
    def kwargs(self):
        return self._kwargs

    @property
    def name(self):
        return self._name

    @property
    def func(self):
        return self._func

    @property
    def X(self):
        return self._X

    @property
    def y(self):
        return self._y

    @property
    def profile_kind(self):
        return self._profile_kind

    @property
    def profile_columns(self):
        return self._profile_columns

    @property
    def input_names(self):
        return []

    @property
    def output_names(self):
        return ""

    @staticmethod
    @abc.abstractmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, t.Any, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, t.Any, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        """supported.
        Given the parameters, is this type of endpoint supported?

        Parameters
        ----------
        X : t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None]
        y : t.Union[pd.Series, pd.DataFrame, np.ndarray, BaseModel, None]
        profile : t.Optional[str]
            the paramters

        Returns
        -------
        bool - true if supported, false else

        """



import shutil

import pytest

from tktl.core.managers.project import ProjectManager


@pytest.fixture(scope="function")
def create_proj():
    ProjectManager.init_project(None, "sample_project")
    yield
    shutil.rmtree("sample_project")



import pandas as pd
import pytest

from tktl.future.registration.validation import validate


@pytest.mark.parametrize(
    "value,sample",
    [
        (
            pd.DataFrame(0, index=range(10000), columns=range(10000)),
            pd.DataFrame(0, index=range(10000), columns=range(10000)),
        )
    ],
)
def test_benchmark_validate(benchmark, value, sample):

    benchmark(validate, value, sample=sample)



import typing as t
from functools import singledispatch

import numpy as np
import pandas as pd  # type: ignore

from .exceptions import ValidationError


@singledispatch
def validate(
    value: t.Union[np.ndarray, pd.DataFrame, pd.Series],
    *,
    sample: t.Union[np.ndarray, pd.DataFrame, pd.Series],
) -> t.Union[np.ndarray, pd.DataFrame, pd.Series]:
    raise ValidationError(f"Can't validate value of type: {type(value)}")


@validate.register
def _validate_numpy(
    value: np.ndarray, *, sample: t.Union[np.ndarray, pd.DataFrame, pd.Series]
) -> np.ndarray:
    if type(value) != type(sample):
        raise ValidationError(
            f"Expected value of type {type(sample)}, got {type(value)}"
        )

    if value.shape[1:] != sample.shape[1:]:
        raise ValidationError(
            "Could not validate numpy array shape. "
            f"Value has shape {value.shape[1:]}, expected shape: {sample.shape[1:]}"
        )

    # We choose not to validate the dtypes here

    return value


@validate.register
def _validate_series(
    value: pd.Series, *, sample: t.Union[np.ndarray, pd.DataFrame, pd.Series]
) -> pd.Series:
    if type(value) != type(sample):
        raise ValidationError(
            f"Expected value of type {type(sample)}, got {type(value)}"
        )

    # We choose not to validate the dtypes here

    return value


@validate.register
def _validate_dataframe(
    value: pd.DataFrame, *, sample: t.Union[np.ndarray, pd.DataFrame, pd.Series]
) -> pd.DataFrame:
    if type(sample) != pd.DataFrame:
        raise ValidationError(
            f"Expected value of type {type(sample)}, got {type(value)}"
        )

    # We choose not to validate the dtypes here

    def decode(x: t.Union[str, bytes]) -> str:
        if isinstance(x, bytes):
            return x.decode()
        return x

    value.columns = [decode(x) for x in value.columns]

    sent_columns = {x for x in value.columns.to_list()}  # type: ignore
    expected_columns = {x for x in sample.columns.to_list()}  # type: ignore

    if sent_columns != expected_columns:
        raise ValidationError(
            "Column mismatch:\n"
            f"Missing columns: {expected_columns.difference(sent_columns)},\n"
            f"Superfluos columns: {sent_columns.difference(expected_columns)}."
        )

    value = value[sample.columns.to_list()]  # type: ignore # ordering
    return value.astype(sample.dtypes.to_dict())  # type: ignore # dtypes. TODO: Improve this



from tktl.core.serializers.arrow.factory import (  # noqa: F401
    deserialize_arrow,
    serialize_arrow,
)
from tktl.core.serializers.rest.factory import (  # noqa: F401
    deserialize_rest,
    serialize_rest,
)
from tktl.core.serializers.rest.typed import to_pydantic  # noqa: F401



import typing as t

import numpy as np

from tktl.future.utils import JSONStructure


def numpy_deserialize(*, value: JSONStructure, sample: np.ndarray) -> np.ndarray:
    return np.array(value)


def numpy_serialize(*, value: np.ndarray) -> JSONStructure:
    return value.tolist()


def numpy_to_example(*, value: np.ndarray) -> t.List[t.Any]:
    return value[0:1].tolist()



import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from pydantic import BaseModel

from tktl.core.future.t import EndpointKinds, ProfileKinds

from .arrow import ArrowEndpoint


class ProfiledEndpoint(ArrowEndpoint):
    kind: EndpointKinds = EndpointKinds.PROFILED

    @property
    def input_names(self):
        return self.profile_columns

    @property
    def output_names(self):
        return self.y.name

    @property
    def profile_columns(self) -> t.Optional[t.List[str]]:
        return (
            self._profile_columns
            if self._profile_columns is not None
            else self.X.columns.to_list()
        )

    @staticmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, t.Any, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, t.Any, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        return (
            profile in ProfileKinds.set()
            and isinstance(X, pd.DataFrame)
            and isinstance(y, pd.Series)
        )



import os

import pytest

from tktl.core.clients.rest import RestClient
from tktl.core.exceptions import TaktileSdkError
from tktl.core.managers.auth import AuthConfigManager


def test_instantiate_client():

    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)

    with pytest.raises(TaktileSdkError):
        RestClient(
            api_key=key,
            repository_name=f"{os.environ['TEST_USER']}/test-new",
            branch_name="master",
            endpoint_name="repayment",
        )

    client = RestClient(
        api_key=key,
        repository_name=f"{os.environ['TEST_USER']}/integ-testing",
        branch_name="master",
        endpoint_name="repayment",
    )
    assert client.location is not None


def test_instantiate_by_url(sample_deployed_url):
    key = os.environ["TEST_USER_API_KEY"]
    AuthConfigManager.set_api_key(key)
    client = RestClient.for_url(
        api_key=key, url=sample_deployed_url, endpoint_name="repayment"
    )
    assert client.location == f"https://{sample_deployed_url}/"



import typing as t

from pydantic import BaseModel

from tktl.core.future.t import ArrowFormatKinds, EndpointKinds, ProfileKinds


class EndpointInfoSchema(BaseModel):
    name: str
    path: str
    kind: EndpointKinds
    profile_kind: t.Optional[ProfileKinds]
    explain_path: t.Optional[str]
    response_kind: t.Optional[ArrowFormatKinds]
    input_names: t.List[str] = []
    output_names: t.Optional[str] = None
    profile_columns: t.Optional[t.List[str]] = None
    input_example: t.Optional[t.List[t.Dict[str, t.Any]]] = None


class InfoEndpointResponseModel(BaseModel):
    schema_version: str
    taktile_cli: str
    profiling: str
    endpoints: t.List[EndpointInfoSchema]



import pytest
from src.endpoints import tktl


@pytest.mark.parametrize(
    "endpoint",
    [
        endpoint
        for endpoint in tktl.endpoints
        if endpoint.input_schema.value is not None
        and not endpoint.disable_rest_validation
    ],
)
def test_endpoints(json_metadata, endpoint):
    """test_endpoints.
    This test ensures the provided sample data on endpoints can be correctly
    processed by the endpoints.

    It is recommended to keep these tests around.
    """

    function = endpoint._func
    function_name = function.__name__

    json_metadata["section"] = "Taktile Automatic Endpoint Tests"
    json_metadata["pass_message"] = f"Sample data for {function_name} is valid"
    json_metadata["fail_message"] = f"Sample data for {function_name} is invalid"

    function(endpoint.input_schema.value)



import click

from tktl.cli.common import (
    ClickGroup,
    get_secrets,
    validate_integration_shared_options,
    validate_project_shared_options,
)
from tktl.commands.validate import (
    validate_all,
    validate_import,
    validate_integration,
    validate_profiling,
    validate_project_config,
    validate_unittest,
)
from tktl.core.config import settings


@click.group(
    "validate", help="Validate the project", cls=ClickGroup, **settings.HELP_COLORS_DICT
)
def validate():
    pass


@validate.command("config", help="Validate the configuration")
@click.option(
    "--path", "-p", help="Validate project located at this path", type=str, default="."
)
def validate_config_command(path) -> None:
    """Validates a new project for the necessary scaffolding, as well as the supporting
    files needed. The directory structure of a new project.
    """
    validate_project_config(path=path)


@validate.command("import", help="Validate src/endpoints.py")
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_import_command(path: str, nocache: bool, secrets_repository: str) -> None:
    secrets = get_secrets(secrets_repository)
    validate_import(path=path, nocache=nocache, secrets=secrets)


@validate.command("unittest", help="Validate the unittests")
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_unittest_command(
    path: str, nocache: bool, secrets_repository: str
) -> None:
    secrets = get_secrets(secrets_repository)
    validate_unittest(path=path, nocache=nocache, secrets=secrets)


@validate.command("integration", help="Validate integration")
@validate_integration_shared_options
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_integration_command(
    path: str, nocache: bool, timeout: int, retries: int, secrets_repository: str
) -> None:
    secrets = get_secrets(secrets_repository)
    validate_integration(
        path=path, nocache=nocache, timeout=timeout, retries=retries, secrets=secrets
    )


@validate.command("profiling", help="Validate profiling")
@validate_integration_shared_options
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_profiling_command(
    path: str, nocache: bool, timeout: int, retries: int, secrets_repository: str
) -> None:
    secrets = get_secrets(secrets_repository)
    validate_profiling(
        path=path, nocache=nocache, timeout=timeout, retries=retries, secrets=secrets
    )


@validate.command("all", help="Validate everything")
@validate_integration_shared_options
@validate_project_shared_options
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
def validate_all_command(
    path: str, nocache: bool, timeout: int, retries: int, secrets_repository: str
) -> None:
    secrets = get_secrets(secrets_repository)
    validate_all(
        path=path, nocache=nocache, timeout=timeout, retries=retries, secrets=secrets
    )



import copy
from json import JSONDecodeError
from typing import List, Optional, Type

import requests
from pydantic import BaseModel, parse_obj_as

from tktl.core import utils
from tktl.core.exceptions.exceptions import APIClientException
from tktl.core.loggers import LOG, Logger
from tktl.core.managers.auth import AuthConfigManager


class API(object):
    DEFAULT_HEADERS = {
        "Accept": "application/json",
    }

    def __init__(self, api_url, headers=None, logger: Logger = LOG):
        """

        Parameters
        ----------
        api_url
        headers
        """

        self.api_url = api_url
        headers = headers or self.DEFAULT_HEADERS
        self.headers = headers.copy()
        self.logger = logger

    @property
    def default_headers(self):
        api_key = AuthConfigManager.get_api_key()
        self.headers.update({"X-Api-Key": api_key})
        return self.headers

    def get_path(self, url=None):
        if not url:
            return self.api_url
        full_path = utils.concatenate_urls(self.api_url, url)
        return full_path

    def post(
        self, url=None, json=None, params=None, files=None, data=None, timeout=None
    ):
        path = self.get_path(url)
        headers = copy.deepcopy(self.default_headers)

        self.logger.trace(
            "POST request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}\n\tfiles: {}\n\tdata: {}".format(
                path, headers, json, params, files, data
            )
        )
        response = requests.post(
            path,
            json=json,
            params=params,
            headers=headers,
            files=files,
            data=data,
            timeout=timeout,
        )
        self.logger.trace("Response status code: {}".format(response.status_code))
        self.logger.trace("Response content: {}".format(response.content))
        return response

    def put(self, url=None, json=None, params=None, data=None):
        path = self.get_path(url)
        self.logger.trace(
            "PUT request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self.headers, json, params
            )
        )
        response = requests.put(
            path, json=json, params=params, headers=self.default_headers, data=data
        )
        self.logger.trace("Response status code: {}".format(response.status_code))
        self.logger.trace("Response content: {}".format(response.content))
        return response

    def patch(self, url=None, json=None, params=None, data=None):
        path = self.get_path(url)
        self.logger.trace(
            "PATCH request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self.headers, json, params
            )
        )
        response = requests.patch(
            path, json=json, params=params, headers=self.default_headers, data=data
        )
        self.logger.trace("Response status code: {}".format(response.status_code))
        self.logger.trace("Response content: {}".format(response.content))
        return response

    def get(self, url=None, json=None, params=None):
        path = self.get_path(url)
        self.logger.trace(
            "GET request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                path, self.headers, json, params
            )
        )
        response = requests.get(
            path, params=params, headers=self.default_headers, json=json
        )
        self.logger.trace("Response status code: {}".format(response.status_code))
        self.logger.trace("Response content: {}".format(response.content))
        return response

    def delete(self, url=None, json=None, params=None):
        path = self.get_path(url)
        response = requests.delete(path, params=params, headers=self.headers, json=json)
        self.logger.trace(
            "DELETE request sent to: {} \n\theaders: {}\n\tjson: {}\n\tparams: {}".format(
                response.url, self.headers, json, params
            )
        )
        self.logger.trace("Response status code: {}".format(response.status_code))
        self.logger.trace("Response content: {}".format(response.content))
        return response


def interpret_response(
    response: requests.Response, model: Optional[Type[BaseModel]], ping: bool = False
):
    """
    Parameters
    ----------
    ping
    response

    model
    """
    try:
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        if e.response.status_code != 500:
            try:
                json_response = response.json()
                raise APIClientException(
                    status_code=e.response.status_code, detail=json_response["detail"]
                )
            except JSONDecodeError:
                raise APIClientException(
                    status_code=e.response.status_code, detail=repr(response)
                )
        else:
            raise APIClientException(
                status_code=e.response.status_code,
                detail=f"Something went wrong: {repr(response.content)}",
            )
    if ping:
        return response
    if model:
        json = response.json()
        if isinstance(json, list):
            if not json:
                return []

            if model.schema()["type"] == "array":
                return model.parse_obj(json)

            elif model.schema()["type"] == "object":
                return parse_obj_as(List[model], json)  # type: ignore
        return model(**response.json())
    else:
        if response.content:
            return response.json()



from taktile_types.enums.instance import ServiceType

from tktl.core.clients import DeploymentApiClient
from tktl.core.clients.arrow import ArrowFlightClient
from tktl.core.clients.rest import RestClient
from tktl.core.clients.taktile import TaktileApiClient
from tktl.core.loggers import LOG, Logger
from tktl.core.managers.auth import AuthConfigManager


class CommandBase(object):
    def __init__(self, api=None):
        self.api = api


class BaseClientCommand:
    def __init__(
        self,
        kind: ServiceType,
        api_key: str,
        repository: str,
        branch_name: str,
        endpoint_name: str = "",
        local: bool = False,
        logger: Logger = LOG,
        skip_auth: bool = False,
    ):
        self.kind = kind
        self.client_params = {
            "api_key": api_key,
            "repository_name": repository,
            "branch_name": branch_name,
            "endpoint_name": endpoint_name,
            "local": local,
            "logger": logger,
            "skip_auth": skip_auth,
        }
        self.client = self._get_client()

    def execute(self, *args, **kwargs):
        pass

    def _get_client(self):
        return (
            RestClient(**self.client_params)
            if self.kind == ServiceType.REST
            else ArrowFlightClient(**self.client_params)
        )


class BaseDeploymentApiCommand:
    def __init__(self, logger: Logger = LOG):
        self.client = DeploymentApiClient(logger=logger)


class BaseTaktileApiCommand:
    def __init__(self, logger: Logger = LOG):
        self.client = TaktileApiClient(logger=logger)


class BaseRestClientCommand(BaseClientCommand):
    def __init__(
        self,
        repository: str,
        branch_name: str,
        local: bool,
        logger: Logger = LOG,
        skip_auth: bool = False,
    ):
        super().__init__(
            kind=ServiceType.REST,
            api_key=AuthConfigManager.get_api_key(),
            repository=repository,
            branch_name=branch_name,
            local=local,
            logger=logger,
            skip_auth=skip_auth,
        )


class BaseGrpcClientCommand(BaseClientCommand):
    def __init__(
        self,
        repository: str,
        branch_name: str,
        local: bool,
        logger: Logger = LOG,
        skip_auth: bool = False,
    ):
        super().__init__(
            kind=ServiceType.GRPC,
            api_key=AuthConfigManager.get_api_key(),
            repository=repository,
            branch_name=branch_name,
            local=local,
            logger=logger,
            skip_auth=skip_auth,
        )



import pytest
from pydantic import UUID4

from tktl.commands.deployments import GetDeployments, GetEndpoints
from tktl.core.exceptions.exceptions import APIClientException
from tktl.login import logout


def test_get_deployment_commands(logged_in_context):
    cmd = GetDeployments()
    result = cmd.execute(
        UUID4("33fb5b8f-e6a2-4dd2-86ef-d257d2c59b85"),
        None,
        None,
        None,
        None,
        None,
        return_all=False,
    )
    assert len(result) == 1

    result = cmd.execute(None, None, None, None, None, None, return_all=True)
    assert len(result) >= 5

    cmd = GetEndpoints()
    result = cmd.execute(
        UUID4("395391a4-a4c8-4671-8dc4-9423fecb5e4a"),
        None,
        None,
        None,
        None,
        None,
        None,
        None,
    )
    assert len(result) == 4

    cmd = GetDeployments()
    with pytest.raises(APIClientException):
        logout()
        cmd.execute(
            UUID4("7c0f6f48-0220-450a-b4d2-bfc731f94cc3"),
            None,
            None,
            None,
            None,
            None,
            return_all=True,
        )



import json
from typing import Dict, List, Tuple

import numpy
import pandas
import pyarrow
import pytest
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal, assert_series_equal

from tktl.core.exceptions import UnsupportedInputTypeException
from tktl.core.serializers.arrow import (
    ArraySerializer,
    BinarySerializer,
    DataFrameSerializer,
    SequenceSerializer,
    SeriesSerializer,
)


def assert_custom_metadata(table, cls_name, deserialize_steps=None):
    assert table.schema.metadata[b"CLS"] == cls_name.encode()
    if deserialize_steps:
        assert (
            json.loads(table.schema.metadata[b"DESERIALIZERS"].decode("utf-8"))
            == deserialize_steps
        )
    if not deserialize_steps and b"DESERIALIZERS" in table.schema.metadata.keys():
        raise ValueError("Should've passed deserialize_steps")


def test_frame_serializer(serializer_df_inputs: pandas.DataFrame):
    as_arrow = DataFrameSerializer.serialize(serializer_df_inputs)
    assert as_arrow.num_rows == len(serializer_df_inputs)
    assert all([a in as_arrow.column_names for a in list(serializer_df_inputs.columns)])
    deserialized = DataFrameSerializer.deserialize(as_arrow)
    assert_frame_equal(serializer_df_inputs, deserialized)
    assert_custom_metadata(table=as_arrow, cls_name=DataFrameSerializer.__name__)


def test_series_serializer(serializer_series_inputs: pandas.Series):
    as_arrow = SeriesSerializer.serialize(serializer_series_inputs)
    assert as_arrow.column(0).null_count == serializer_series_inputs.isna().sum()
    assert len(as_arrow) == len(serializer_series_inputs)
    deserialized = SeriesSerializer.deserialize(as_arrow)
    assert_series_equal(serializer_series_inputs, deserialized)
    assert_custom_metadata(table=as_arrow, cls_name=SeriesSerializer.__name__)


def test_series_serializer_multi_index(serializer_series_inputs: pandas.Series):
    serializer_series_inputs.index = pandas.MultiIndex.from_arrays(
        [
            serializer_series_inputs.index.tolist(),
            numpy.random.randn(len(serializer_series_inputs)).tolist(),
        ]
    )
    as_arrow = SeriesSerializer.serialize(serializer_series_inputs)
    deserialized = SeriesSerializer.deserialize(as_arrow)
    assert_series_equal(serializer_series_inputs, deserialized)


def test_frame_serializer_with_custom_metadata(serializer_df_inputs: pandas.DataFrame):
    # https://arrow.apache.org/docs/python/data.html#custom-schema-and-field-metadata
    # Undocumented as of 28/9/2020
    as_arrow = DataFrameSerializer.serialize(serializer_df_inputs)
    assert as_arrow.num_rows == len(serializer_df_inputs)
    assert all([a in as_arrow.column_names for a in list(serializer_df_inputs.columns)])
    deserialized = DataFrameSerializer.deserialize(as_arrow)
    assert_frame_equal(serializer_df_inputs, deserialized)


def test_sequence_serializer(
    serializer_dict_inputs: Dict,
    serializer_l_of_d_sequence_inputs: List[Dict],
    serializer_list_and_tuple_inputs: Tuple[List, List, Tuple],
):
    ser = SequenceSerializer.serialize(serializer_dict_inputs)
    assert isinstance(ser, pyarrow.Table)
    deser = SequenceSerializer.deserialize(ser)
    assert isinstance(deser, dict)
    assert deser.keys() == serializer_dict_inputs.keys()
    assert deser["a"] == serializer_dict_inputs["a"]
    assert deser["b"] == serializer_dict_inputs["b"]
    assert_custom_metadata(
        table=ser,
        cls_name=SequenceSerializer.__name__,
        deserialize_steps=["NATIVE_dict", "FINAL_first"],
    )

    ser = SequenceSerializer.serialize(serializer_l_of_d_sequence_inputs)
    assert isinstance(ser, pyarrow.Table)
    deser = SequenceSerializer.deserialize(ser)
    assert isinstance(deser, list)
    assert len(deser) == len(serializer_l_of_d_sequence_inputs)
    assert all(
        [deser[i] == serializer_l_of_d_sequence_inputs[i] for i in range(len(deser))]
    )

    expected_pipeline_steps = (
        ["SELECT_column", "NATIVE_list"],
        ["SELECT_column", "NATIVE_list"],
        ["SELECT_column", "NATIVE_list"],
        ["SELECT_column", "NATIVE_list"],
        ["SELECT_column", "NATIVE_list", "FINAL_tuple"],
    )

    for inp, eps in zip(serializer_list_and_tuple_inputs, expected_pipeline_steps):
        ser = SequenceSerializer.serialize(inp)
        assert isinstance(ser, pyarrow.Table)
        deser = SequenceSerializer.deserialize(ser)
        assert deser == inp
        assert_custom_metadata(
            table=ser, cls_name=SequenceSerializer.__name__, deserialize_steps=eps
        )


def test_1d_and_2d_array_serializer():
    for shapes in [(100,), (50, 2), (25, 4), (10, 10)]:
        arr = numpy.random.randn(100).reshape(shapes)
        as_arrow = ArraySerializer.serialize(arr)
        assert isinstance(as_arrow, pyarrow.Table)
        deserialized = ArraySerializer.deserialize(as_arrow)
        assert isinstance(deserialized, numpy.ndarray)
        assert deserialized.shape == arr.shape
        assert_array_equal(deserialized, arr)
        assert_custom_metadata(table=as_arrow, cls_name=ArraySerializer.__name__)


def test_numpy_ndim_gt_2_arbitrary_serializer():
    arr = numpy.random.randn(100).reshape((10, 5, 2))
    with pytest.raises(UnsupportedInputTypeException):
        ArraySerializer.serialize(arr)

    as_arrow = BinarySerializer.serialize(arr)
    assert isinstance(as_arrow, pyarrow.Table)
    deserialized = BinarySerializer.deserialize(as_arrow)
    assert isinstance(deserialized, numpy.ndarray)
    assert deserialized.shape == arr.shape
    assert_array_equal(deserialized, arr)



import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore
from pydantic import BaseModel

from tktl.core.future.t import EndpointKinds

from .advanced import AdvancedEndpoint


class TypedEndpoint(AdvancedEndpoint):
    kind: EndpointKinds = EndpointKinds.TYPED

    @staticmethod
    def supported(
        *,
        X: t.Union[pd.Series, pd.DataFrame, np.ndarray, t.Any, BaseModel, None] = None,
        y: t.Union[pd.Series, pd.DataFrame, np.ndarray, t.Any, BaseModel, None] = None,
        profile: t.Optional[str] = None,
    ) -> bool:
        return (
            profile is None
            and isinstance(X, (type(BaseModel), type(t.Any)))
            and isinstance(y, (type(BaseModel), type(t.Any)))
        )



import json
import os

from click.testing import CliRunner

from tktl import main


def test_login(user_key):
    """Test the CLI."""
    u, k = user_key
    runner = CliRunner()
    result = runner.invoke(main.login, [k])
    print(result.output)
    assert result.exit_code == 0
    assert result.output == f"Authentication successful for user: {u}\n"

    result = runner.invoke(main.login, ["FAKEAPIKEY"])
    assert result.exit_code == 0
    assert result.output == "Authentication failed: Key format is invalid\n"
    with open(os.path.expanduser("~/.config/tktl/config.json"), "r") as j:
        d = json.load(j)
    assert d["api-key"] != "FAKEAPIKEY"
    assert d["api-key"] == k



import math
import typing as t
from http import HTTPStatus

import pandas as pd  # type: ignore
import requests
import tenacity  # type: ignore
from pyarrow import Table  # type: ignore
from pyarrow.flight import FlightClient  # type: ignore
from pyarrow.flight import ActionType, FlightDescriptor
from requests import ConnectionError, Timeout

from tktl.core.clients.future.http import API
from tktl.core.config import settings
from tktl.core.exceptions.exceptions import (
    APIClientException,
    APIClientExceptionRetryable,
)
from tktl.core.loggers import LOG
from tktl.core.serializers import deserialize_arrow, serialize_arrow

# codes for which retry-ing is not necessary, and for which
# precise error information can be retrieved either from
# the request, or deployment logs
DO_NOT_RETRY_STATUS_CODES = [
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.FORBIDDEN,
    HTTPStatus.UNAUTHORIZED,
    HTTPStatus.UNPROCESSABLE_ENTITY,
]


def safe_post(
    *, api: API, url: str, data: t.Any, retries: int = 3, timeout: float = 10.0
):
    """safe_post.
    HTTP POST request with retrying

    Parameters
    ----------
    api : API
        the underlying api
    url : str
        url to make the request to
    data : t.Any
        data to be sent
    retries : int
        retries
    timeout : float
        timeout
    """

    def my_after(retry_state):
        LOG.warning(
            f"Timeout while trying to call endpoint on try #{retry_state.attempt_number}/{retries}: {retry_state.outcome}"
        )

    def my_stop(retry_state):
        if retry_state.attempt_number >= retries:
            LOG.error(
                f"Giving up trying to call endpoint after {retry_state.attempt_number} attempts"
            )
            return True
        return False

    @tenacity.retry(
        stop=my_stop,
        retry=tenacity.retry_if_exception_type(
            exception_types=(APIClientExceptionRetryable,)
        ),
        wait=tenacity.wait_random(min=0, max=1),
        after=my_after,
        reraise=True,
    )
    def wrapped():
        try:
            response = api.post(url=url, data=data, timeout=timeout)
        except (Timeout, ConnectionError) as e:
            raise APIClientExceptionRetryable(
                status_code=-1, detail="A connection error has occured"
            ) from e

        if response.status_code in DO_NOT_RETRY_STATUS_CODES:
            if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
                raise APIClientException(
                    status_code=response.status_code,
                    detail="An error occurred with your request. Check the deployment's logs for more information",
                )
            elif response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
                json_response = response.json()
                raise APIClientException(
                    status_code=response.status_code,
                    detail=f"Unprocessable request body: {json_response['detail']})",
                )
            else:
                json_response = response.json()
                raise APIClientException(
                    status_code=response.status_code,
                    detail=f"Authentication Error: {json_response['detail']}",
                )
        try:
            response.raise_for_status()
        except requests.exceptions.RequestException as e:
            try:
                json_response = response.json()
                raise APIClientExceptionRetryable(
                    status_code=response.status_code, detail=json_response["detail"],
                ) from e
            except (ValueError, KeyError) as e:
                # Note that we use ValueError instead of JSONDecodeError above on purpose,
                # because depending on runtime configuration `requests` might choose to
                # use `simplejson` which has a different JSONDecodeError. However both derive
                # from ValueError.
                content = response.text
                raise APIClientExceptionRetryable(
                    status_code=response.status_code,
                    detail=f"An error occurred with your request: {content})",
                ) from e
        return response.json()

    return wrapped()


def batch_arrow(
    *, client: FlightClient, action: ActionType, X: t.Any, use_input_index: bool = False
):
    table = serialize_arrow(X)
    batch_size, batch_memory = _get_chunk_size(table)
    if not (batch_size and batch_memory):
        return
    descriptor = client.get_flight_info(FlightDescriptor.for_command(action.type))
    writer, reader = client.do_exchange(descriptor.descriptor)  # type: ignore
    LOG.trace(
        f"Initiating prediction request with batches of {batch_size} records of "
        f"~{batch_memory:.2f} MB/batch"
    )
    batches = table.to_batches(max_chunksize=batch_size)
    chunks = []
    schema = None
    with writer:
        writer.begin(table.schema)
        for i, batch in enumerate(batches):
            LOG.trace(f"Prediction for batch {i + 1}/{len(batches)}")
            chunk = _send_batch(
                writer=writer, batch=batch, reader=reader, batch_number=i + 1
            )
            if not chunk:
                continue
            if not schema and chunk.data.schema is not None:
                schema = chunk.data.schema
            chunks.append(chunk.data)
    deserialized = deserialize_arrow(Table.from_batches(chunks, schema))
    if use_input_index:
        input_has_index = isinstance(X, pd.Series) or isinstance(X, pd.DataFrame)
        output_has_index = isinstance(deserialized, pd.Series) or isinstance(
            deserialized, pd.DataFrame
        )
        if not input_has_index or not output_has_index:
            LOG.warning(
                "Inputs or Outputs are not of type series or dataframe, use_input_index has no effect"
            )
        else:
            try:
                deserialized.index = X.index
            except Exception as e:
                LOG.warning(f"Unable to set indexes of output frame: {repr(e)}")
    return deserialized


def _send_batch(writer, batch, reader, batch_number):
    try:
        writer.write_batch(batch)
        return reader.read_chunk()
    except Exception as e:
        LOG.error(
            f"ERROR: performing prediction for batch {batch_number}: {e} "
            f"The predictions from this batch will be missing from the result"
        )
        return None


def _get_chunk_size(sample_table: Table) -> t.Tuple[t.Optional[int], t.Optional[float]]:
    try:
        mem_per_record = sample_table.nbytes / sample_table.num_rows
    except ZeroDivisionError:
        LOG.warning("Empty dataframe received")
        return None, None
    batch_size = math.ceil(settings.ARROW_BATCH_MB * 1e6 / mem_per_record)
    batch_memory_mb = (batch_size * mem_per_record) / 1e6
    return batch_size, batch_memory_mb



from collections.abc import Sequence
from functools import singledispatch
from typing import List, Type, Union, _GenericAlias  # type: ignore

import numpy  # type: ignore
import pandas  # type: ignore
from pydantic import BaseModel

from tktl.core.serializers import rest
from tktl.core.serializers.base import CustomDeserializingModelT


def deserialize_rest(
    model_input: Union[List[CustomDeserializingModelT], CustomDeserializingModelT]
):
    if isinstance(model_input, list):
        return [v.deserialize() for v in model_input]
    else:
        return model_input.deserialize()


@singledispatch
def serialize_rest(
    model_input, output_model: Union[Type[BaseModel], Type[CustomDeserializingModelT]]
):
    return output_model(**{"value": model_input})


@serialize_rest.register
def _pass_through(model_input: BaseModel, output_model: Type[BaseModel]):
    return rest.PassThroughSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register
def _sequence(model_input: Sequence, output_model: Type[BaseModel]):
    return rest.SequenceSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register(dict)
def _dict(model_input, output_model: Type[BaseModel]):
    return rest.SequenceSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register
def _ndarray(model_input: numpy.ndarray, output_model: Type[BaseModel]):
    return rest.ArraySerializer.serialize(value=model_input, output_model=output_model)


@serialize_rest.register
def _df(model_input: pandas.DataFrame, output_model: Type[BaseModel]):
    return rest.DataFrameSerializer.serialize(
        value=model_input, output_model=output_model
    )


@serialize_rest.register
def _series(model_input: pandas.Series, output_model: Type[BaseModel]):
    return rest.SeriesSerializer.serialize(value=model_input, output_model=output_model)


@serialize_rest.register
def _generic(model_input: _GenericAlias, output_model: Type[BaseModel]):
    return model_input



import json
import os
from http import HTTPStatus
from typing import Optional

import requests
import tenacity  # type: ignore
from openapi_schema_pydantic import OpenAPI  # type: ignore
from requests import ConnectionError, Timeout
from taktile_types.enums.instance import ServiceType

from tktl.core.clients import Client
from tktl.core.clients.http_client import API, interpret_response
from tktl.core.config import settings
from tktl.core.exceptions.exceptions import (
    APIClientException,
    APIClientExceptionRetryable,
)
from tktl.core.loggers import LOG, Logger
from tktl.core.schemas.repository import _format_http_url

# codes for which retry-ing is not necessary, and for which
# precise error information can be retrieved either from
# the request, or deployment logs
DO_NOT_RETRY_STATUS_CODES = [
    HTTPStatus.INTERNAL_SERVER_ERROR,
    HTTPStatus.FORBIDDEN,
    HTTPStatus.UNAUTHORIZED,
    HTTPStatus.UNPROCESSABLE_ENTITY,
]


class RestClient(Client):
    def get_schema(self, *args, **kwargs):
        pass

    TRANSPORT = ServiceType.REST

    def __init__(
        self,
        api_key: str,
        repository_name: str,
        branch_name: str,
        endpoint_name: str,
        local: bool = False,
        logger: Logger = LOG,
        verbosity: int = 0,
        skip_auth: bool = False,
    ):
        super().__init__(
            api_key=api_key,
            repository_name=repository_name,
            branch_name=branch_name,
            endpoint_name=endpoint_name,
            local=local,
            logger=logger,
            verbosity=verbosity,
        )
        if not skip_auth:
            self._authenticate()

    def predict(self, inputs, retries=3, timeout=10):
        """predict.
        Use model endpoint

        Parameters
        ----------
        inputs :
            inputs variables
        retries : int
            number of retries
        timeout : float
            timeout in seconds

        Raises
        ------
        requests.Timeout
            If the last attempt resulted in a timeout exception
        requests.ConnectionError
            If the last attempt was not able to connect
        EndpointException
            If the last attempt resulted in a connection error
        """

        def my_after(retry_state):
            self.logger.warning(
                f"Timeout while trying to predict on try #{retry_state.attempt_number}/{retries}: {retry_state.outcome}"
            )

        def my_stop(retry_state):
            if retry_state.attempt_number >= retries:
                self.logger.error(
                    f"Giving up trying to predict after {retry_state.attempt_number} attempts"
                )
                return True
            return False

        @tenacity.retry(
            stop=my_stop,
            retry=tenacity.retry_if_exception_type(
                exception_types=(APIClientExceptionRetryable,)
            ),
            wait=tenacity.wait_random(min=0, max=1),
            after=my_after,
            reraise=True,
        )
        def wrapped():
            try:
                response = self.client.post(
                    url=f"model/{self.endpoint_name}",
                    data=json.dumps(inputs),
                    timeout=timeout,
                )
            except (Timeout, ConnectionError) as e:
                raise APIClientExceptionRetryable(
                    status_code=-1, detail="A connection error has occured"
                ) from e

            if response.status_code in DO_NOT_RETRY_STATUS_CODES:
                if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
                    raise APIClientException(
                        status_code=response.status_code,
                        detail="An error occurred with your request. Check the deployment's logs for more information",
                    )
                elif response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
                    json_response = response.json()
                    raise APIClientException(
                        status_code=response.status_code,
                        detail=f"Unprocessable request body: {json_response['detail']})",
                    )
                else:
                    json_response = response.json()
                    raise APIClientException(
                        status_code=response.status_code,
                        detail=f"Authentication Error: {json_response['detail']}",
                    )
            try:
                response.raise_for_status()
            except requests.exceptions.RequestException as e:
                try:
                    json_response = response.json()
                    raise APIClientExceptionRetryable(
                        status_code=response.status_code,
                        detail=json_response["detail"],
                    ) from e
                except (ValueError, KeyError) as e:
                    # Note that we use ValueError instead of JSONDecodeError above on purpose,
                    # because depending on runtime configuration `requests` might choose to
                    # use `simplejson` which has a different JSONDecodeError. However both derive
                    # from ValueError.
                    content = response.text
                    raise APIClientExceptionRetryable(
                        status_code=response.status_code,
                        detail=f"An error occurred with your request: {content})",
                    ) from e
            return response.json()

        return wrapped()

    @classmethod
    def for_url(cls, api_key: str, url: str, endpoint_name: str):
        rest_url = cls.format_url(url=url)
        instantiated = cls(
            api_key=api_key,
            repository_name="",
            branch_name="",
            endpoint_name=endpoint_name,
            skip_auth=True,
        )
        client = API(api_url=rest_url)
        instantiated.set_client_and_location(location=rest_url, client=client)
        instantiated.logger.trace(f"Performing authentication request against {url}")
        try:
            instantiated.health()
        except APIClientException as e:
            instantiated.logger.error(f"Could not instantiate client: {e.detail}")
            return
        return instantiated

    @staticmethod
    def format_url(url: Optional[str]) -> str:
        return _format_http_url(url, docs=False)

    def _authenticate(self, health_check: bool = False):
        if health_check:
            location = self.get_deployment_location()
        else:
            location = self.get_endpoint_and_location()
        client = API(api_url=location)
        self.set_client_and_location(location=location, client=client)

    @property
    def local_endpoint(self):
        return settings.LOCAL_REST_ENDPOINT

    def list_deployments(self):
        pass

    def get_sample_data(self):
        if not self.endpoint_has_rest_sample_data:
            self.logger.warning("Sample data not available for this endpoint")
            return None, None
        schema = self.client.get(url="openapi.json")
        openapi = OpenAPI.parse_obj(schema.json())
        request_reference, response_reference = get_endpoint_model_reference(
            openapi, endpoint=self.endpoint_name
        )
        sample_input = openapi.components.schemas[request_reference].example
        if not sample_input:
            self.logger.warning("No sample input found for endpoint")
        sample_output = openapi.components.schemas[response_reference].example
        if not sample_output:
            self.logger.warning("No sample output found for endpoint")
        return sample_input, sample_output

    def health(self):
        response = self.client.get(url="healthz")
        return interpret_response(response, model=None, ping=True)


def get_endpoint_model_reference(openapi: OpenAPI, endpoint: str):
    for k in openapi.paths.keys():
        if f"/model/{endpoint}" == k:
            request_ref = (
                openapi.paths[k]
                .post.requestBody.content["application/json"]
                .media_type_schema.ref
            )
            response_ref = (
                openapi.paths[k]
                .post.responses["200"]
                .content["application/json"]
                .media_type_schema.ref
            )
            return os.path.basename(request_ref), os.path.basename(response_ref)
    return None



import joblib
import pandas as pd

from tktl import Tktl

# instantiate client
tktl = Tktl()

# load model
model = joblib.load("assets/model.joblib")

# load reference data
data = pd.read_parquet("assets/loans_test.pqt")
label = "Repaid"
X = data.drop(columns=label)
y = data[label]


# specify transformation
@tktl.endpoint(kind="binary", X=X, y=y)
def repayment(df):
    pred = model.predict_proba(df)[:, 1]
    return pred



import click
import click_completion  # type: ignore

from tktl.cli.common import ClickGroup
from tktl.cli.delete import delete as delete_commands
from tktl.cli.get import get as get_commands
from tktl.cli.health import health
from tktl.cli.init import init
from tktl.cli.login import login, logout
from tktl.cli.run import run
from tktl.cli.validate import validate
from tktl.commands.version import get_version
from tktl.core.config import settings
from tktl.core.loggers import set_verbosity

click_completion.init()


@click.group(cls=ClickGroup, **settings.HELP_COLORS_DICT)
@click.option("-v", "--verbose", count=True, help="Set verbosity level")
@click.option("--future", default=False, is_flag=True, help="Use future implementation")
def cli(verbose, future):
    set_verbosity(verbose)
    settings.FUTURE_ENDPOINTS = future


cli.add_command(get_version)
cli.add_command(logout)
cli.add_command(login)
cli.add_command(init)
cli.add_command(get_version)
cli.add_command(delete_commands)
cli.add_command(get_commands)
cli.add_command(validate)
cli.add_command(run)
cli.add_command(health)
